///////////////////////////////////////////////////////////////////////
// FLV plugin for VirtualDub 1.8.8+
// Copyright (C) 2010-2012 fccHandler
// Copyright (C) 1998-2012 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

// IVDXAudioSource implementation

#include <windows.h>
#include <vfw.h>
#include <stdio.h>
#include <crtdbg.h>

#include "vd2\plugin\vdinputdriver.h"
#include "Unknown.h"

#include "WinFile.h"
#include "DataVector.h"

#include "FLVTypes.h"
#include "InputFileFLV.h"
#include "AudioSourceFLV.h"


// ADPCM samples per block?
#define FLV_ADPCM_SAMPLES 1024

// MPEG Layer III search size
#define MP3_SEARCH_SIZE 2048


/**********************************************************************
 *                                                                    *
 *                      AudioTrackFLV                                 *
 *                                                                    *
 **********************************************************************/

AudioTrackFLV::AudioTrackFLV()
	: dv                (NULL)
	, pHeader           (NULL)
	, iHeaderLen        (0)
	, pFormat           (NULL)
	, iFormatLen        (0)
	, sample_list       (NULL)
	, dwLength          (0)
	, dwMaxSample       (0)
	, dwTotalBytes      (0)
	, Duration          (0)
	, dwRate            (1)
	, dwScale           (1)
	, bFormat           (255)
	, bFormatHi         (255)
	, bHasVSB           (false)
	, bDeliverRawAAC    (false)
{
}


AudioTrackFLV::~AudioTrackFLV()
{
	delete[] pFormat;
	delete[] sample_list;
	delete[] pHeader;
	delete dv;
}


bool AudioTrackFLV::ParseTag(WinFile &mFile, const FLVTag &tag)
{
	FLVSampleInfo fsi;
	unsigned char fmt;

	bool ret = false;

	fsi.pos  = mFile.Pos();
	fsi.size = tag.DataSize[0];
	fsi.size = (fsi.size << 8) + tag.DataSize[1];
	fsi.size = (fsi.size << 8) + tag.DataSize[2];

	// Zero size payload is not an error
	if (fsi.size <= 0)
	{
		ret = true;
		goto Abort;
	}

	fsi.pres_time = tag.TimestampExtended;
	fsi.pres_time = (fsi.pres_time << 8) + tag.Timestamp[0];
	fsi.pres_time = (fsi.pres_time << 8) + tag.Timestamp[1];
	fsi.pres_time = (fsi.pres_time << 8) + tag.Timestamp[2];

	// First, check if format is supported and
	// if so determine the details

	fmt = (unsigned char)mFile.NextByte();
	++fsi.pos;
	--fsi.size;

	// Zero size payload is not an error
	if (fsi.size <= 0)
	{
		ret = true;
		goto Abort;
	}

	if (bFormatHi == 255)
	{
		// First audio tag, save the format
		bFormat = fmt;
		bFormatHi = bFormat >> 4;
	}
	else
	{
		// Subsequent audio tag, verify the format
		if (fmt != bFormat)
		{
			// This should never happen, but if it does
			// just ignore it; do not abort the load.
			ret = true;
			goto Abort;
		}
	}

	switch (bFormatHi) {

		case FLV_AUDIO_PCM:		// Linear PCM, platform endian
			goto Abort;

		case FLV_AUDIO_ADPCM:	// ADPCM

			if (dv == NULL)
			{
				dv = new DataVector(sizeof(FLVSampleInfo));
				if (dv == NULL) goto Abort;
			}

			dv->Add(&fsi);
			break;

		case FLV_AUDIO_MP3:		// MP3

			if (dv == NULL)
			{
				dv = new DataVector(sizeof(FLVSampleInfo));
				if (dv == NULL) goto Abort;
			}

			if (pHeader == NULL)
			{
				// Get the first MP3 header
				// (Overrules the tag format)

				pHeader = new unsigned char[MP3_SEARCH_SIZE];
				if (pHeader == NULL) goto Abort;
			}

			if (iHeaderLen < MP3_SEARCH_SIZE)
			{
				int remains = fsi.size;

				if (remains > (MP3_SEARCH_SIZE - iHeaderLen))
					remains =  MP3_SEARCH_SIZE - iHeaderLen;

				mFile.Read(pHeader + iHeaderLen, remains);
				iHeaderLen += remains;
			}

			dv->Add(&fsi);
			break;

		case FLV_AUDIO_PCMLE:	// Linear PCM, little endian
		case FLV_AUDIO_NELLY16:	// Nellymoser 16-kHz mono
		case FLV_AUDIO_NELLY8:	// Nellymoser 8-kHz mono
		case FLV_AUDIO_NELLY:	// Nellymoser
		case FLV_AUDIO_ALAW:	// G.711 A-law logarithmic PCM
		case FLV_AUDIO_ULAW:	// G.711 mu-law logarithmic PCM
		case 9:					// Reserved
			goto Abort;

		case FLV_AUDIO_AAC:		// AAC

			++fsi.pos;
			--fsi.size;

			if (mFile.NextByte() == 0 && fsi.size >= 2)
			{
				// AudioSpecificConfig, at least 2 bytes
				// (Overrules the tag format)

				if (pHeader == NULL)
				{
					// Needs to fit in cbSize field!
					if (fsi.size > 65535) goto Abort;

					pHeader = new unsigned char[fsi.size];
					if (pHeader == NULL) goto Abort;
					iHeaderLen = fsi.size;
					
					mFile.Read(pHeader, fsi.size);
				}
			}
			else if (fsi.size != 0 && iHeaderLen > 0)
			{
				// Raw data
				if (dv == NULL)
				{
					dv = new DataVector(sizeof(FLVSampleInfo));
					if (dv == NULL) goto Abort;
				}
				dv->Add(&fsi);
			}
			break;

		case FLV_AUDIO_SPEEX:	// Speex
		case 12:				// Reserved
		case 13:				// Reserved
		case 14:				// MP3 8-Khz
		case 15:				// Device-specific sound
			goto Abort;
	}

	ret = true;

Abort:
	return ret;
}


bool AudioTrackFLV::FormatADPCM()
{
	static const unsigned short freq_map[4] = { 5513, 11025, 22050, 44100 };
	WAVEFORMATEX *wfex;

	if (bHasVSB) return false;	// Variable blocksize not supported

	// xxxx....  format
	// ....xx..  sampling rate (0=5.5, 1=11, 2=22, 3=44 KHz)
	// ......x.  bits per sample (0=8, 1=16)
	// .......x  channels (0=1, 1=2)

	if ((bFormat & 0x0C) == 0) goto Abort;	// 5.5 KHz not supported
	if ((bFormat & 0x02) == 0) goto Abort;	// 8-bit not supported

	pFormat = new unsigned char[sizeof(WAVEFORMATEX)];
	if (pFormat == NULL) goto Abort;
	iFormatLen = sizeof(WAVEFORMATEX);

	wfex = (WAVEFORMATEX *)pFormat;

	wfex->wFormatTag      = WAVE_FORMAT_PCM;
	wfex->nChannels       = (bFormat & 1)? 2: 1;
	wfex->nSamplesPerSec  = freq_map[(bFormat >> 2) & 3];
	wfex->wBitsPerSample  = (bFormat & 2)? 16: 8;
	wfex->nBlockAlign     = wfex->nChannels * (wfex->wBitsPerSample / 8);
	wfex->nAvgBytesPerSec = wfex->nSamplesPerSec * wfex->nBlockAlign;
	wfex->cbSize          = 0;

	dwRate  = wfex->nSamplesPerSec;
	dwScale = 1;

	return true;
	
Abort:
	return false;
}


bool AudioTrackFLV::FormatMP3()
{
	static const unsigned char chan_map[4] = { 2, 2, 2, 1 };

	static const unsigned short freq_map[4] = { 11025, 12000, 8000, 0 };

	MPEGLAYER3WAVEFORMAT *af;
	unsigned long hdr;
	int spf;
	int i;
	
	if (pHeader == NULL || iHeaderLen < 4) goto Abort;

	// Locate the first MP3 header.
	hdr = 0;
	for (i = 0; i < iHeaderLen; ++i)
	{
		hdr = (hdr << 8) + pHeader[i];

		if ((hdr & 0xFFE00000) == 0xFFE00000 &&
		    (hdr & 0x00060000) == 0x00020000 &&
		    (hdr & 0x0000F000) != 0x0000F000 &&
		    (hdr & 0x00000C00) != 0x00000C00 )
		{
			break;
		}
	}
	if (i == iHeaderLen) goto Abort;

// 11111111 1111.... ........ ........   sync (all bits set)
// ........ ....x... ........ ........   version (1=MPEG-1, 0=MPEG-2)
// ........ .....xx. ........ ........   layer (00=reserved,11=I,10=II,01=III)
// ........ .......x ........ ........   protected by CRC (0=yes, 1=no)
// ........ ........ xxxx.... ........   bitrate_index (can't be 1111)
// ........ ........ ....xx.. ........   sampling_freq (can't be 11)
// ........ ........ ......x. ........   padding_bit (0=no, 1=yes)
// ........ ........ .......x ........   private_bit (not used)
// ........ ........ ........ xx......   (00=stereo,10=joint,01=dual,11=mono)
// ........ ........ ........ ..xx....   mode_ext (only if joint stereo)
// ........ ........ ........ ....x...   copyright_bit (0=no, 1=yes)
// ........ ........ ........ .....x..   original_or_copy (0=original, 1=copy)
// ........ ........ ........ ......xx   emphasis (00=none,01=reserved,10=50/15ms,11=CCIT J.17)

	pFormat = new unsigned char[sizeof(MPEGLAYER3WAVEFORMAT)];
	if (pFormat == NULL) goto Abort;
	iFormatLen = sizeof(MPEGLAYER3WAVEFORMAT);

	af = (MPEGLAYER3WAVEFORMAT *)pFormat;

	af->wfx.wFormatTag      = WAVE_FORMAT_MPEGLAYER3;
	af->wfx.nChannels       = chan_map[(hdr >>  6) & 3];
	af->wfx.nSamplesPerSec  = freq_map[(hdr >> 10) & 3];
	af->wfx.nAvgBytesPerSec = (DWORD)((double)dwTotalBytes / Duration + 0.5);
	af->wfx.nBlockAlign     = (WORD)dwMaxSample;
	af->wfx.wBitsPerSample  = 0;
	af->wfx.cbSize          = MPEGLAYER3_WFX_EXTRA_BYTES;

	spf = 1152;		// samples per frame
	if (hdr & 0x00100000)   { af->wfx.nSamplesPerSec *= 2; spf /= 2; }
	if (hdr & 0x00080000)   { af->wfx.nSamplesPerSec *= 2; spf /= 2; }

	af->wID                 = MPEGLAYER3_ID_MPEG;
	af->fdwFlags            = MPEGLAYER3_FLAG_PADDING_ISO;
	af->nBlockSize          = (WORD)dwMaxSample;
	af->nFramesPerBlock     = 1;
	af->nCodecDelay         = 0;

	dwRate  = af->wfx.nSamplesPerSec;
	dwScale = spf;

	return true;
	
Abort:
	return false;
}


bool AudioTrackFLV::ParseASC()
{
	static const unsigned char chan_map[16] = {
		0, 1, 2, 3, 4, 5, 6, 8, 0, 0, 0, 0, 0, 0, 0, 0
	};

	static const unsigned long freq_map[16] = {
		96000, 88200, 64000, 48000, 44100, 32000, 24000, 22050,
		16000, 12000, 11025,  8000,  7350,     0,     0,     0
	};

	WAVEFORMATEX *wfex;

	if (pFormat == NULL) goto Abort;
	if (pHeader == NULL) goto Abort;
	if (iHeaderLen < 2) goto Abort;
	
	wfex = (WAVEFORMATEX *)pFormat;

	// pHeader is AudioSpecificConfig:

	// 00: 00010...  object type (00010 = LC)
	//     .....xxx  sampling frequency index (0 to 11)
	// 01: x.......  sampling frequency index continued
	//     .xxxx...  channels configuration (1 to 7)
	//     .....x..  frame length flag (0 = 1024, 1 = 960)
	//     ......0.  depends on core coder (0)
	//     .......0  extensions flag (0)

	wfex->nSamplesPerSec = freq_map[((pHeader[0] << 1) | (pHeader[1] >> 7)) & 0xF];
	if (wfex->nSamplesPerSec == 0) goto Abort;

	wfex->nChannels = chan_map[(pHeader[1] >> 3) & 0xF];
	if (wfex->nChannels == 0) goto Abort;

	// Detect explicit SBR and/or PS

	// 02: 01010110  syncExtensionType (0x02B7)
	// 03: 111.....  syncExtensionType continued
	//     ...00101  extensionAudioObjectType (5)
	// 04: 1.......  sbrPresentFlag (0 = no, 1 = yes)
	//     .xxxx...  extensionSamplingFrequencyIndex
	//     .....101  syncExtensionType (0x0548)
	// 05: 01001000  syncExtensionType continued
	// 06: x.......  psPresentFlag (0 = no, 1 = yes)

	if (   iHeaderLen >= 5
	   &&  pHeader[2] == 0x56
	   &&  pHeader[3] == 0xE5
	   )
	{
		if (pHeader[4] & 0x80) wfex->nSamplesPerSec *= 2;			

		if (   iHeaderLen >= 7
		   && (pHeader[4] & 0x07) == 0x05
		   &&  pHeader[5] == 0x48
		   )
		{
			if (pHeader[6] & 0x80) ++wfex->nChannels;
		}
	}
	else
	{
		if (wfex->nSamplesPerSec <= 24000)
		{
			wfex->nSamplesPerSec *= 2;
		}
	}

	return true;

Abort:
	return false;
}


bool AudioTrackFLV::FormatAAC()
{
	static const unsigned char chan_map[16] = {
		0, 1, 2, 3, 4, 5, 6, 8, 0, 0, 0, 0, 0, 0, 0, 0
	};

	static const unsigned long freq_map[16] = {
		96000, 88200, 64000, 48000, 44100, 32000, 24000, 22050,
		16000, 12000, 11025,  8000,  7350,     0,     0,     0
	};

	WAVEFORMATEX *wfex;
	DWORD dwSamplesPerPacket;

	if (pHeader == NULL || iHeaderLen < 2) goto Abort;

	// pHeader is AudioSpecificConfig:
	//   LSB      MSB
	// xxxxx... ........  object type (00010 = AAC LC)
	// .....xxx x.......  sampling rate
	// ........ .xxxx...  channel configuration
	// ........ .....x..  frame length flag (0 = 1024, 1 = 960)
	// ........ ......x.  depends on core coder (0)
	// ........ .......x  extension flag (0)

	// Hatsune has 13 90:
	// 00010... ........  object type (00010 = AAC LC)
	// .....011 1.......  sampling rate
	// ........ .0010...  channel configuration

	pFormat = new unsigned char[sizeof(WAVEFORMATEX) + iHeaderLen];
	if (pFormat == NULL) goto Abort;
	iFormatLen = sizeof(WAVEFORMATEX) + iHeaderLen;

	// Derive channels and sampling rate from AudioSpecificConfig	
	if (!ParseASC()) goto Abort;

	// Samples per packet
	dwSamplesPerPacket = (pHeader[1] & 4)? 960: 1024;

	wfex = (WAVEFORMATEX *)pFormat;

	wfex->wFormatTag      = WAVE_FORMAT_RAW_AAC1;
	wfex->nAvgBytesPerSec = (DWORD)((double)dwTotalBytes / Duration + 0.5);
	wfex->nBlockAlign     = (WORD)dwMaxSample;
	wfex->wBitsPerSample  = 0;
	wfex->cbSize          = iHeaderLen;

	memcpy(pFormat + sizeof(WAVEFORMATEX), pHeader, iHeaderLen);

	dwRate  = wfex->nSamplesPerSec;
	dwScale = dwSamplesPerPacket;

	return true;

Abort:
	return false;
}



bool AudioTrackFLV::Finish()
{
	unsigned long i;

	if (dv == NULL) goto Abort;

	sample_list = (FLVSampleInfo *)dv->MakeArray();
	dwLength = dv->Length();

	delete dv;
	dv = NULL;

	if (dwLength == 0) goto Abort;

	dwTotalBytes = 0;
	dwMaxSample  = sample_list[0].size;

	for (i = 0; i < dwLength; ++i)
	{
		unsigned long size = sample_list[i].size;
		if (size != dwMaxSample)
		{
			bHasVSB = true;
			if (size > dwMaxSample) dwMaxSample = size;
		}
		dwTotalBytes += size;
	}

	if (bFormatHi == FLV_AUDIO_AAC && !bDeliverRawAAC)
	{
		// Adjust for ADTS headers
		dwMaxSample  += 7;
		dwTotalBytes += 7 * dwLength;
	}

	// Calculate duration in seconds
	Duration = 1.0;		// in case we fail

	// We must have at least two samples.
	if (dwLength > 1)
	{
		DWORD dwLo  = 0;
		DWORD dwHi  = 0;
		DWORD dwMin = sample_list[0].pres_time;
		DWORD dwMax = dwMin;
		
		for (i = 1; i < dwLength; ++i)
		{
			if (sample_list[i].pres_time < dwMin)
			{
				dwMin = sample_list[i].pres_time;
				dwLo  = i;
			}
			if (sample_list[i].pres_time > dwMax)
			{
				dwMax = sample_list[i].pres_time;
				dwHi  = i;
			}
		}

		// (last - first) / (numframes - 1) = milliseconds per frame
		// duration = (milliseconds per frame * numframes) / 1000

		if (dwHi > dwLo)
		{
			// milliseconds per frame:
			Duration = ((double)dwMax - (double)dwMin) / ((double)dwHi - (double)dwLo);
			Duration = Duration * dwLength * 0.001;
		}
	}

	switch (bFormatHi) {

	case FLV_AUDIO_ADPCM:
		if (!FormatADPCM()) goto Abort;
		break;

	case FLV_AUDIO_MP3:
		if (!FormatMP3()) goto Abort;
		break;

	case FLV_AUDIO_AAC:
		if (!FormatAAC()) goto Abort;
		break;

	default:
		goto Abort;
	}

	return true;
	
Abort:
	return false;
}


/**********************************************************************
 *                                                                    *
 *                      AudioSourceFLV                                *
 *                                                                    *
 **********************************************************************/

AudioSourceFLV::AudioSourceFLV(InputFileFLV *const pp, AudioTrackFLV *const at)
	: parentPtr(pp)
	, audio(at)
{
	parentPtr->AddRef();
	errmode = kErrorModeDecodeAnyway;

	if (audio->bFormatHi == FLV_AUDIO_ADPCM)
	{
		const WAVEFORMATEX *wfex = (WAVEFORMATEX *)audio->pFormat;
		
		const unsigned int i = (audio->dwMaxSample + 15) & 0xFFFFFFF0;
		const unsigned int framesize = 512 * (wfex->nSamplesPerSec / 11025);

		adpcm_buf = new unsigned char[i + framesize * wfex->nBlockAlign];
		if (adpcm_buf != NULL) adpcm_out = adpcm_buf + i;

	}
	else
	{
		adpcm_buf = NULL;
	}

_RPT0(_CRT_WARN, "AudioSourceFLV constructed\n");
}


AudioSourceFLV::~AudioSourceFLV()
{
	delete[] adpcm_buf;
	parentPtr->Release();

_RPT0(_CRT_WARN, "AudioSourceFLV destructed\n");
}


int AudioSourceFLV::AddRef() {
_RPT0(_CRT_WARN, "AudioSourceFLV::AddRef\n");
	return vdxunknown<IVDXStreamSource>::AddRef();
}


int AudioSourceFLV::Release() {
_RPT0(_CRT_WARN, "AudioSourceFLV::Release\n");
	return vdxunknown<IVDXStreamSource>::Release();
}


void *VDXAPIENTRY AudioSourceFLV::AsInterface(uint32 iid) {
	if (iid == IVDXAudioSource::kIID)
		return static_cast<IVDXAudioSource *>(this);

	if (iid == IVDXStreamSourceV3::kIID)
		return static_cast<IVDXStreamSourceV3 *>(this);

	return vdxunknown<IVDXStreamSource>::AsInterface(iid);
}


void VDXAPIENTRY AudioSourceFLV::GetStreamSourceInfo(VDXStreamSourceInfo& srcInfo)
{
	switch (audio->bFormatHi) {

	case FLV_AUDIO_ADPCM:
	{
		const WAVEFORMATEX *wfex = (WAVEFORMATEX *)audio->pFormat;
		const unsigned int framesize = 512 * (wfex->nSamplesPerSec / 11025);

		srcInfo.mSampleRate.mNumerator   = audio->dwRate;
		srcInfo.mSampleRate.mDenominator = audio->dwScale;
		srcInfo.mSampleCount             = audio->dwLength * framesize;
		break;
	}

	case FLV_AUDIO_MP3:
	case FLV_AUDIO_AAC:
	default:
		srcInfo.mSampleRate.mNumerator   = audio->dwRate;
		srcInfo.mSampleRate.mDenominator = audio->dwScale;
		srcInfo.mSampleCount             = audio->dwLength;
		break;
	}

	srcInfo.mPixelAspectRatio.mNumerator   = 1;
	srcInfo.mPixelAspectRatio.mDenominator = 1;
}


void VDXAPIENTRY AudioSourceFLV::GetStreamSourceInfoV3(VDXStreamSourceInfoV3& info)
{
	GetStreamSourceInfo(info.mInfo);

	if (audio->bHasVSB)
		info.mFlags = VDXStreamSourceInfoV3::kFlagVariableSizeSamples;
	else
		info.mFlags = 0;

	info.mfccHandler = 0;
}


int AudioSourceFLV::adpcm_bits(int nBits)
{
	unsigned long cur_ptr = adpcm_ptr >> 3;
	int ret = 0;

	if (cur_ptr < adpcm_len)
	{
		int curbyte = adpcm_buf[cur_ptr++];
		int curbit = 128U >> (adpcm_ptr & 7);

		while (nBits > 0)
		{
			ret += ret;
			if (curbyte & curbit) ++ret;

			++adpcm_ptr;
			--nBits;
			curbit >>= 1;

			if (curbit == 0)
			{
				if (cur_ptr >= adpcm_len) break;
				curbyte = adpcm_buf[cur_ptr++];
				curbit = 128;
			}
		}
	}
	return ret;
}


void AudioSourceFLV::adpcm_decode()
{
	// Decode "framesize" samples from adpcm_buf to adpcm_out.

	const WAVEFORMATEX *wfex = (WAVEFORMATEX *)audio->pFormat;
	const unsigned int framesize = 512 * (wfex->nSamplesPerSec / 11025);

	short *dst = (short *)adpcm_out;

	static const int step_table[89] = {
		7, 8, 9, 10, 11, 12, 13, 14, 16, 17,
		19, 21, 23, 25, 28, 31, 34, 37, 41, 45,
		50, 55, 60, 66, 73, 80, 88, 97, 107, 118,
		130, 143, 157, 173, 190, 209, 230, 253, 279, 307,
		337, 371, 408, 449, 494, 544, 598, 658, 724, 796,
		876, 963, 1060, 1166, 1282, 1411, 1552, 1707, 1878, 2066,
		2272, 2499, 2749, 3024, 3327, 3660, 4026, 4428, 4871, 5358,
		5894, 6484, 7132, 7845, 8630, 9493, 10442, 11487, 12635, 13899,
		15289, 16818, 18500, 20350, 22385, 24623, 27086, 29794, 32767
	};

	static const int swf_index_tables[4][16] = {
		/*2*/ { -1, 2 },
		/*3*/ { -1, -1, 2, 4 },
		/*4*/ { -1, -1, -1, -1, 2, 4, 6, 8 },
		/*5*/ { -1, -1, -1, -1, -1, -1, -1, -1, 1, 2, 4, 6, 8, 10, 13, 16 }
	};

	int predictor[2];
	int step_index[2];

	const int *table;
	unsigned int k0, count;
    int signmask, nb_bits;
	int i;

	// adpcm_len must be initialized first!
	adpcm_ptr = 0;

	//read bits & initial values
	nb_bits = adpcm_bits(2) + 2;
	table = swf_index_tables[nb_bits - 2];
	k0 = 1U << (nb_bits - 2);
	signmask = 1 << (nb_bits - 1);

	for (i = 0; i < wfex->nChannels; ++i)
	{
		predictor[i] = adpcm_bits(16);
		if (predictor[i] & 0x8000) predictor[i] |= 0xFFFF0000;
		*dst++ = (short)predictor[i];
		step_index[i] = adpcm_bits(6);
	}

    for (count = 1; count < framesize; ++count)
	{
        for (i = 0; i < wfex->nChannels; ++i)
		{
            // similar to IMA adpcm
            int delta = adpcm_bits(nb_bits);
            int step = step_table[step_index[i]];
            int vpdiff = 0;
            unsigned int k = k0;

            do {
                if (delta & k) vpdiff += step;
                step >>= 1;
                k >>= 1;
            } while (k);

            vpdiff += step;

            if (delta & signmask)
                predictor[i] -= vpdiff;
            else
                predictor[i] += vpdiff;

            step_index[i] += table[delta & (~signmask)];

			if (step_index[i] < 0)
				step_index[i] = 0;
			else if (step_index[i] > 88)
				step_index[i] = 88;

			if (predictor[i] < -32768)
				predictor[i] = -32768;
			else if (predictor[i] > 32767)
				predictor[i] = 32767;

            *dst++ = (short)predictor[i];
        }
    }
}


bool AudioSourceFLV::Read(sint64 lStart64, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead)
{
	const WAVEFORMATEX *wfex = (WAVEFORMATEX *)audio->pFormat;

	DWORD dwStart = (DWORD)lStart64;

	if (audio->bFormatHi == FLV_AUDIO_ADPCM)
	{
		const unsigned int framesize = 512 * (wfex->nSamplesPerSec / 11025);

		DWORD adpcmStart = dwStart / framesize;
		DWORD adpcmMod   = dwStart % framesize;
		DWORD actual;

		const DWORD ch = wfex->nChannels;
		const DWORD ba = wfex->nBlockAlign;

		// buffer size query or buffer too small?
		if (lpBuffer == NULL || cbBuffer < ba)
		{
			actual = lCount * ba;
			if (actual < ba) actual = ba;

			if (lBytesRead) *lBytesRead = actual;
			if (lSamplesRead) *lSamplesRead = actual / ba;
			return (lpBuffer == NULL);
		}

		if (adpcmStart >= audio->dwLength || lCount == 0)
		{
			if (lBytesRead) *lBytesRead = 0;
			if (lSamplesRead) *lSamplesRead = 0;
			return true;
		}

		actual = 0;
		do {
			adpcm_len = parentPtr->ReadData(adpcm_buf, audio->sample_list[adpcmStart].pos, audio->sample_list[adpcmStart].size);
			adpcm_decode();

			uint32 tc = framesize - adpcmMod;
			if (tc > lCount) tc = lCount;

			memcpy(lpBuffer, adpcm_out + adpcmMod * ba, tc * ba);

			lCount -= tc;
			actual += tc;
			if (lCount == 0) break;
			lpBuffer = (char *)lpBuffer + tc * ba;

			++adpcmStart;
			adpcmMod = 0;
		} while (adpcmStart < audio->dwLength);

		if (lBytesRead) *lBytesRead = actual * ba;
		if (lSamplesRead) *lSamplesRead = actual;
	}
	else
	{
		const FLVSampleInfo *fsi;
		unsigned char *dst;
		DWORD actual;
		DWORD i;
		
		unsigned int profile;
		unsigned int chcode;
		unsigned int srcode;
		unsigned int iADTSlen;

		// Start out of range?
		if (dwStart >= audio->dwLength)
		{
			if (lBytesRead) *lBytesRead = 0;
			if (lSamplesRead) *lSamplesRead = 0;
			return true;
		}

		// Sample count out of range?
		if (lCount > (audio->dwLength - dwStart))
		{
			lCount = audio->dwLength - dwStart;
		}

		// Wrap AAC frames in ADTS headers?
		if (audio->bFormatHi == FLV_AUDIO_AAC && !audio->bDeliverRawAAC)
		{
			iADTSlen = 7;
		}
		else
		{
			iADTSlen = 0;
		}

		// Buffer size query?
		if (lpBuffer == NULL)
		{
			fsi = &(audio->sample_list[dwStart]);
			actual = fsi->size + iADTSlen;
			++fsi;

			for (i = 1; i < lCount; ++i)
			{
				actual += fsi->size + iADTSlen;
				++fsi;
			}

			if (actual < audio->dwMaxSample)
			{
				actual = audio->dwMaxSample;
			}

			if (lBytesRead) *lBytesRead = actual;
			if (lSamplesRead) *lSamplesRead = i;

			return true;
		}

		// Buffer too small to hold a sample?
		if (cbBuffer < audio->dwMaxSample)
		{
			if (lBytesRead) *lBytesRead = audio->dwMaxSample;
			if (lSamplesRead) *lSamplesRead = 1;
			return false;
		}

		if (iADTSlen)
		{
			// Gather information for ADTS header	

			// AudioSpecificConfig:
			//   LSB      MSB
			// xxxxx... ........  object type (00010 = AAC LC)
			// .....xxx x.......  sampling rate
			// ........ .xxxx...  channel configuration
			// ........ .....x..  frame length flag (0 = 1024, 1 = 960)
			// ........ ......x.  depends on core coder (0)
			// ........ .......x  extension flag (0)

			profile = *((WORD *)(audio->pFormat + sizeof(WAVEFORMATEX)));
			profile = ((profile & 0xFF) << 8) | (profile >> 8);

			chcode  = (profile >> 3) & 7;
			srcode  = (profile >> 7) & 15;
			profile = ((profile >> 11) - 1) & 3;
		}

		// Begin reading samples
		dst = (unsigned char *)lpBuffer;
		fsi = &(audio->sample_list[dwStart]);

		for (i = 0; i < lCount; ++i)
		{
			if (cbBuffer < (fsi->size + iADTSlen)) break;

			if (iADTSlen)
			{
				// Build ADTS header:
				// 00: 11111111   syncword
				// 01: 1111....   syncword (continued)
				//     ....0...   id (1 = MPEG-2, 0 = MPEG-4)
				//     .....00.   layer
				//     .......1   protection absent
				// 02: xx......   profile
				//     ..xxxx..   sf_index
				//     ......0.   private bit
				//     .......x   channel configuration
				// 03: xx......   channel configuration (continued)
				//     ..0.....   original
				//     ...0....   home
				//     ....0...   copyright identification bit
				//     .....0..   copyright identification start
				//     ......xx   frame length
				// 04: xxxxxxxx   frame length (continued)
				// 05: xxx.....   frame length (continued)
				//     ...11111   buffer fullness
				// 06: 111111..   buffer fullness (continued)
				//     ......00   raw data blocks in frame

				unsigned int frlen = (fsi->size + 7) & 0x1FFF;
			
				dst[0] = 0xFF;
				dst[1] = 0xF1;
				dst[2] = (unsigned char)((profile << 6) | (srcode << 2) | (chcode >> 2));
				dst[3] = (unsigned char)(((chcode & 3) << 6) | ((frlen >> 11) & 3));
				dst[4] = (unsigned char)((frlen >> 3) & 0xFF);
				dst[5] = (unsigned char)(((frlen & 7) << 5) | 0x1F);
				dst[6] = 0xFC;
			}

			actual = parentPtr->ReadData(dst + iADTSlen, fsi->pos, fsi->size);
			if (actual != fsi->size) break;
			dst += iADTSlen + actual;
			cbBuffer -= actual;
			++fsi;
		}

		if (lBytesRead) *lBytesRead = (uint32)(dst - (unsigned char *)lpBuffer);
		if (lSamplesRead) *lSamplesRead = i;
	}

	return true;
}


const void *AudioSourceFLV::GetDirectFormat()
{
	return audio->pFormat;
}


int AudioSourceFLV::GetDirectFormatLen()
{
	return audio->iFormatLen;
}


IVDXStreamSource::ErrorMode AudioSourceFLV::GetDecodeErrorMode()
{
	return errmode;
}


void AudioSourceFLV::SetDecodeErrorMode(IVDXStreamSource::ErrorMode mode)
{
	errmode = mode;
}


bool AudioSourceFLV::IsDecodeErrorModeSupported(IVDXStreamSource::ErrorMode mode)
{
	return true;
}


bool AudioSourceFLV::IsVBR()
{
	return false;
}


sint64 AudioSourceFLV::TimeToPositionVBR(sint64 us)
{
	return 0;
}


sint64 AudioSourceFLV::PositionToTimeVBR(sint64 samples)
{
	return 0;
}


void AudioSourceFLV::GetAudioSourceInfo(VDXAudioSourceInfo& info)
{
	info.mFlags = 0;
}

