///////////////////////////////////////////////////////////////////////
// FLV plugin for VirtualDub 1.8.8+
// Copyright (C) 2010-2012 fccHandler
// Copyright (C) 1998-2012 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

// IVDXVideoDecoder, IVDXVideoDecoderModel,
// and IVDXVideoSource implementations

#if defined(_MSC_VER) && (_MSC_VER >= 1400)
	#define _CRT_SECURE_NO_WARNINGS
#endif

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <vfw.h>
#include <stdio.h>


// Use VirtualDub's I/P decoder model?
#define USE_DEFAULT_IP_MODEL

// Defer video decoding to VirtualDub?
//#define DEFER_DECODING

// Workaround for variable frame rate?
//#define FLV_VFR


#include "vd2\plugin\vdinputdriver.h"
#include "Unknown.h"

#include "WinFile.h"
#include "FLVTypes.h"

#include "DataVector.h"
#include "VideoCodec.h"

#include "InputFileFLV.h"
#include "VideoSourceFLV.h"
#include "AVCSequence.h"

#include <crtdbg.h>


// Maximum supported width and height of a frame
#define FLV_MAX_WIDTH  1920
#define FLV_MAX_HEIGHT 1088


///////////////////////////////////////////////////////////////////////
//
//                      VideoDecoderFLV
//
///////////////////////////////////////////////////////////////////////


class VideoDecoderFLV : public vdxunknown<IVDXVideoDecoder> {

public:
	VideoDecoderFLV(VideoSourceFLV *const pp, VideoTrackFLV *const vt);
	~VideoDecoderFLV();

	// IVDXVideoDecoder
	const void*      VDXAPIENTRY DecodeFrame(const void *inputBuffer, uint32 data_len, bool is_preroll, sint64 sampleNumber, sint64 targetFrame);
	uint32           VDXAPIENTRY GetDecodePadding();
	void             VDXAPIENTRY Reset();
	bool             VDXAPIENTRY IsFrameBufferValid();
	const VDXPixmap& VDXAPIENTRY GetFrameBuffer();
	bool             VDXAPIENTRY SetTargetFormat(int format, bool useDIBAlignment);
	bool             VDXAPIENTRY SetDecompressedFormat(const VDXBITMAPINFOHEADER *pbih);
	bool             VDXAPIENTRY IsDecodable(sint64 sample_num);
	const void*      VDXAPIENTRY GetFrameBufferBase();

private:
	VideoSourceFLV*  const parentPtr;
	VideoTrackFLV*   const video;

	VideoCodec       mCodec;

	sint64           mPrevFrame;
	unsigned long    dwActualFrame;
	unsigned long    dwGuess;
};


VideoDecoderFLV::VideoDecoderFLV(VideoSourceFLV *const pp, VideoTrackFLV *const vt)
	: parentPtr(pp)
	, video(vt)
{
	parentPtr->AddRef();
	video->hCodec = NULL;

	dwGuess = 0;

	dwActualFrame = -1;
	mPrevFrame = -1;
}


VideoDecoderFLV::~VideoDecoderFLV()
{
	video->hCodec = NULL;
	parentPtr->Release();
}


const void *VideoDecoderFLV::DecodeFrame(
	const void *inputBuffer, uint32 data_len, bool is_preroll,
	sint64 streamFrame,	sint64 targetFrame)
{

//_RPT4(_CRT_WARN, "DecodeFrame %08X, %i, %I64i, %I64i\n", (int)inputBuffer,
//	(int)is_preroll, streamFrame, targetFrame);

	DWORD dwFrame;
	
	if (streamFrame < 0) streamFrame = targetFrame;

	if ((unsigned long)streamFrame >= video->dwPerceivedLength) {
		return mCodec.mFrameBuffer;
	}

	dwFrame = (DWORD)streamFrame;

	// If dwFrame is the same as dwActualFrame,
	// then we've already decoded this frame.

	if (dwFrame != dwActualFrame)
	{
		DWORD f = video->sample_list[dwFrame].size;

		dwActualFrame = dwFrame;

		if ((f & FLV_KEYFRAME) == 0) {
			f = ICDECOMPRESS_NOTKEYFRAME;
		} else {
			f = 0;
		}

		if (is_preroll) {
			f |= ICDECOMPRESS_PREROLL;
		} else {
			mPrevFrame = streamFrame;
		}

		mCodec.Decompress((LPVOID)inputBuffer, data_len, f);
	}

	return mCodec.mFrameBuffer;
}


uint32 VideoDecoderFLV::GetDecodePadding()
{
	return 0;
}


void VideoDecoderFLV::Reset()
{
	dwActualFrame = -1;
	mPrevFrame = -1;
}


bool VideoDecoderFLV::IsFrameBufferValid()
{
	return (mPrevFrame >= 0);
}


const VDXPixmap& VideoDecoderFLV::GetFrameBuffer()
{
	return mCodec.mPixmap;
}


bool VideoDecoderFLV::SetTargetFormat(int format, bool useDIBAlignment)
{
	char buf[64];
	unsigned char fcc[5];
	int i;

	Reset();

	if (mCodec.Init(&video->bmih, format))
	{
		video->hCodec = mCodec.hCodec;
		return true;
	}

	if (mCodec.hCodec != NULL) return false;
	video->hCodec = NULL;

	*(unsigned long *)fcc = video->bmih.biCompression;
	for (i = 0; i < 4; ++i)
	{
		if (fcc[i] < 32) fcc[i] = '?';
	}
	fcc[4] = 0;
	_snprintf(buf, sizeof(buf), "MISSING CODEC (%s)", (char *)fcc);
	buf[sizeof(buf) - 1] = '\0';

	// Missing codec
	mCodec.Inform(buf);

	return (format == mCodec.mPixmap.format ||
		format == nsVDXPixmap::kPixFormat_Null);
}


bool VideoDecoderFLV::SetDecompressedFormat(const VDXBITMAPINFOHEADER *pbih)
{
	return false;
}


const void *VideoDecoderFLV::GetFrameBufferBase()
{
	return mCodec.mFrameBuffer;
}


bool VideoDecoderFLV::IsDecodable(sint64 sample_num64)
{
	if ((DWORD)sample_num64 < video->dwPerceivedLength) {
		return (sample_num64 == (mPrevFrame + 1));
	}
	return false;
}


#ifndef USE_DEFAULT_IP_MODEL

///////////////////////////////////////////////////////////////////////
//
//                      VideoDecoderModelFLV
//
///////////////////////////////////////////////////////////////////////

class VideoDecoderModelFLV : public vdxunknown<IVDXVideoDecoderModel> {

public:
	VideoDecoderModelFLV(VideoSourceFLV *const pp, VideoTrackFLV *const vt);
	~VideoDecoderModelFLV();

	// IVDXVideoDecoderModel
	void    VDXAPIENTRY Reset();
	void    VDXAPIENTRY SetDesiredFrame(sint64 frame_num);
	sint64  VDXAPIENTRY GetNextRequiredSample(bool& is_preroll);
	int     VDXAPIENTRY GetRequiredCount();

private:
	VideoSourceFLV* const parentPtr;
	VideoTrackFLV*  const video;

	sint64          mLastFrame;
	sint64          mTargetFrame;
	sint64          mNextFrame;
	DWORD           dwGuess;

	bool            IsKey(sint64 sample);
};


VideoDecoderModelFLV::VideoDecoderModelFLV(VideoSourceFLV *const pp, VideoTrackFLV *const vt)
	: parentPtr(pp)
	, video(vt)
{
	parentPtr->AddRef();
	mLastFrame = -1;
	dwGuess = 0;
}


VideoDecoderModelFLV::~VideoDecoderModelFLV()
{
	parentPtr->Release();
}


void VideoDecoderModelFLV::Reset()
{
	mLastFrame = -1;
}


bool VideoDecoderModelFLV::IsKey(sint64 sample)
{
	return video->IsKey((unsigned long)sample);
}


void VideoDecoderModelFLV::SetDesiredFrame(sint64 frame_num)
{
	mTargetFrame = frame_num;

	// Do we already have the frame?
	if (mLastFrame == frame_num) {
		mNextFrame = -1;	// Flag this for GetNextRequiredSample
		return;
	}

	mNextFrame = frame_num;

	// If it's a Delta frame, locate a previous Key
	if (!IsKey(frame_num)) {

		while (mNextFrame > 0) {
			--mNextFrame;
			if (mNextFrame == mLastFrame) {
				// Start from the frame following that one
				++mNextFrame;
				break;
			}
			if (IsKey(mNextFrame)) break;
		}
	}
}


sint64 VideoDecoderModelFLV::GetNextRequiredSample(bool& is_preroll)
{
	sint64 frame;

	if (mNextFrame == mTargetFrame || mNextFrame < 0) {
		is_preroll = false;
		mLastFrame = mTargetFrame;
		frame = mNextFrame;
		mNextFrame = -1;
	} else {
		is_preroll = true;
		frame = mNextFrame++;
	}

	return frame;
}


int VideoDecoderModelFLV::GetRequiredCount()
{
	if (mNextFrame < 0) return 0;

	return (int)((mTargetFrame - mNextFrame) + 1);
}

#endif	// USE_DEFAULT_IP_MODEL


///////////////////////////////////////////////////////////////////////
//
//                      VideoTrackFLV
//
///////////////////////////////////////////////////////////////////////


VideoTrackFLV::VideoTrackFLV()
	: dv                (NULL)

	, sample_list       (NULL)
	, sample_map        (NULL)
	, dwActualLength    (0)
	, dwPerceivedLength (0)

	, bFormat           (255)
	, hCodec            (NULL)

	, dwRate            (1)
	, dwScale           (1)

	, avc_dcr           (NULL)
	, dcr_size          (0)
	, avc_len           (0)
	, avc_w             (0)
	, avc_h             (0)
{
	memset(&bmih, 0, sizeof(bmih));
}


VideoTrackFLV::~VideoTrackFLV()
{
	delete[] sample_list;
	delete[] avc_dcr;
	delete dv;
}


int VideoTrackFLV::TransformAVCConfRec(const unsigned char *src, unsigned char *dst)
{
	// Transform the AVCDecoderConfigurationRecord
	// into something that ffdshow can understand
	// by unwrapping it into raw NALUs.  Return the
	// number of bytes in the result.  We have to
	// save this data and prepend it to all key
	// frames of the AVC stream.

	int n;
	int total = 0;

	++src;						// configurationVersion (1)
	++src;						// AVCProfileIndication
	++src;						// profile_compatibility
	++src;						// AVCLevelIndication

	avc_len = (*src++ & 3) + 1;	// '111111', lengthSizeMinusOne
	n = *src++ & 0x1F;			// numOfSequenceParameterSets

	while (n-- > 0)
	{
		int size = *src++;
		size = (size << 8) + *src++;

		if (size > 0)
		{
			if (dst != NULL)
			{
				*dst++ = 0;
				*dst++ = 0;
				*dst++ = 0;
				*dst++ = 1;

				memcpy(dst, src, size);
				dst += size;
			}

			src   += size;
			total += 4 + size;
		}
	}

	n = *src++;					// numOfPictureParameterSets

	while (n-- > 0)
	{
		int size = *src++;
		size = (size << 8) + *src++;

		if (size > 0)
		{
			if (dst != NULL)
			{
				*dst++ = 0;
				*dst++ = 0;
				*dst++ = 0;
				*dst++ = 1;

				memcpy(dst, src, size);
				dst += size;
			}

			src   += size;
			total += 4 + size;
		}
	}

	return total;
}


bool VideoTrackFLV::ParseTag(WinFile &mFile, const FLVTag &tag)
{
	FLVSampleInfo fsi;
	DWORD fourcc;
	unsigned char fmt;
	//int format;
	int w, h;
	bool key;
	bool ret;

	ret = false;

	fsi.pos = mFile.Pos();

	fsi.size = tag.DataSize[0];
	fsi.size = (fsi.size << 8) + tag.DataSize[1];
	fsi.size = (fsi.size << 8) + tag.DataSize[2];

	// Zero size payload is not an error
	if (fsi.size <= 0)
	{
		ret = true;
		goto Abort;
	}

	// First, check if format is supported and
	// if so determine the video width and height.

	fmt = (unsigned char)mFile.NextByte();
	++fsi.pos;
	--fsi.size;

	// Zero size payload is not an error
	if (fsi.size <= 0)
	{
		ret = true;
		goto Abort;
	}

	key = ((fmt >> 4) == 1);
	fmt &= 15;

	if (bFormat == 255)
	{
		// First video tag, save the format
		bFormat = fmt;
	}
	else
	{
		// Subsequent video tag, verify the format
		if (fmt != bFormat)
		{
			// This should never happen, but if it does
			// just ignore it; do not abort the load.
			ret = true;
			goto Abort;
		}
	}


	switch (bFormat) {

		case 0:					// Reserved
		case FLV_VIDEO_JPEG:	// JPEG
			goto Abort;

		case FLV_VIDEO_H263:	// Sorenson H.263
			fourcc = '1VLF';

// 00000000 00000000 1....... ........ ........ picture_start_code
// ........ ........ .xxxxx.. ........ ........ format (00000 or 00001)
// ........ ........ ......xx xxxxxx.. ........ temporal_reference
// ........ ........ ........ ......xx x....... type_information

			mFile.AlignToByte();
			if (mFile.GetBits(17) != 1) goto Abort;

			w = mFile.GetBits(5);	// format
			if (w != 0 && w != 1) goto Abort;

			mFile.GetBits(8);		// temporal_reference
			w = mFile.GetBits(3);	// type_information

			switch (w) {

				case 0:
					w = mFile.GetBits(8);
					h = mFile.GetBits(8);
					break;

				case 1:
					w = mFile.GetBits(16);
					h = mFile.GetBits(16);
					break;

				case 2:
					w = 352; h = 288;
					break;

				case 3:
					w = 176; h = 144;
					break;

				case 4:
					w = 128; h = 96;
					break;

				case 5:
					w = 320; h = 240;
					break;

				case 6:
					w = 160; h = 120;
					break;

				case 7:
					goto Abort;
			}
			break;

		case FLV_VIDEO_SCREEN:	// Screen video
			goto Abort;

		case FLV_VIDEO_VP6:		// On2 VP6
			fourcc = 'F6PV';

			mFile.NextByte();	// unknown
			++fsi.pos;
			--fsi.size;

			goto On2VP6;

		case FLV_VIDEO_VP6A:	// On2 VP6 with alpha channel
			fourcc = 'F6PV';	// (Do not use 'A6PV', it doesn't work!)

			// I do not see this in MPlayer,
			// but it was in Moitah's version:

			mFile.Seek(mFile.Pos() + 4);
			fsi.pos  += 4;
			fsi.size -= 4;

On2VP6:

	// x....... ........  key_frame (0=yes,1=no)
	// .xxxxxx. ........  quantizer
	// .......x ........  separated_coeff
	// ........ xxxxx...  sub_version
	// ........ .....xx.  filter_header

			mFile.AlignToByte();

			key = (mFile.NextBit() == 0);

			if (!key) {
				// Delta frames do not have width/height
				w = bmih.biWidth;
				h = bmih.biHeight;
				break;
			}

			mFile.GetBits(6);			// quantizer
			w = mFile.NextBit();		// separated_coeff
			mFile.GetBits(5);			// sub_version
			h = mFile.GetBits(2);		// filter_header
			mFile.NextBit();			// interlace

			if (w != 0 || h == 0)
				mFile.GetBits(16);		// coeff_offset

			h = mFile.GetBits(8) * 16;	// macroblock_rows
			w = mFile.GetBits(8) * 16;	// macroblock_cols

			// The code below handles the special case of
			// VP6 files which do not start with a keyframe.

			// In this case, we have kept the orphan delta frames
			// in order to preserve A/V sync, but the BM width and
			// height fields were set to zero.  Now that we have
			// a keyframe, we can set the actual width and height.

			if (w == 0 || w > FLV_MAX_WIDTH) goto Abort;			
			if (h == 0 || h > FLV_MAX_HEIGHT) goto Abort;			
			
			if (bmih.biWidth == 0 && bmih.biHeight == 0)
			{
				bmih.biWidth = w;
				bmih.biHeight = h;
				bmih.biSizeImage = ((w * 3 + 3) & -4) * h;
			}

			break;

		case FLV_VIDEO_SCREEN2:	// Screen video version 2
			goto Abort;

		case FLV_VIDEO_AVC:		// AVC

			fourcc = '462H';
			fourcc = '1CVA';

			w = mFile.NextByte();
			mFile.Seek(mFile.Pos() + 3);
			fsi.pos  += 4;
			fsi.size -= 4;

			if (fsi.size <= 0)
			{
				ret = true;
				goto Abort;
			}

			if (w == 0)
			{
				AVCSequence seq;
				unsigned char *tmp;

				// Get AVCDecoderConfigurationRecord
				tmp = new unsigned char[fsi.size];
				if (tmp == NULL) goto Abort;
				mFile.Read(tmp, fsi.size);

				// First, get the size needed
				dcr_size = TransformAVCConfRec(tmp, NULL);
				avc_dcr  = new unsigned char[dcr_size + 1];
				dcr_size = TransformAVCConfRec(tmp, avc_dcr);
				delete[] tmp;

				if (!seq.ParseSPS(avc_dcr + 4, dcr_size - 4)) goto Abort;
				avc_w = seq.Width();
				avc_h = seq.Height();

				if (avc_dcr != NULL && dcr_size > 0) ret = true;
				goto Abort;
			}

			if (w != 1) {
				ret = true;
				goto Abort;
			}

			w = avc_w;
			h = avc_h;
			break;

		default:	// Reserved
			goto Abort;
	}

	// OK, create new video array if applicable,
	// otherwise check new format against current

	if (dv == NULL)
	{
		// Width and height supported?
		if (w < 0 || w > FLV_MAX_WIDTH) goto Abort;
		if (h < 0 || h > FLV_MAX_HEIGHT) goto Abort;
		
		if (w == 0 || h == 0)
		{
			// Probably a VP6 file which doesn't begin with a keyframe.
			if (fourcc != 'F6PV' && fourcc != 'A6PV') goto Abort;
		}

		dv = new DataVector(sizeof(FLVSampleInfo));
		if (dv == NULL) goto Abort;

		bmih.biSize = sizeof(bmih);
		bmih.biWidth = w;
		bmih.biHeight = h;
		bmih.biPlanes = 1;
		bmih.biCompression = fourcc;
		bmih.biBitCount = 24;
		bmih.biSizeImage = ((bmih.biWidth * 3 + 3) & -4) * bmih.biHeight;
		bmih.biXPelsPerMeter = 0;
		bmih.biYPelsPerMeter = 0;
		bmih.biClrUsed = 0;
		bmih.biClrImportant = 0;
	}
	else
	{
		if (w != bmih.biWidth) goto Abort;
		if (h != bmih.biHeight) goto Abort;
	}

	// All good, finish the sample info

	fsi.pres_time = tag.TimestampExtended;
	fsi.pres_time = (fsi.pres_time << 8) + tag.Timestamp[0];
	fsi.pres_time = (fsi.pres_time << 8) + tag.Timestamp[1];
	fsi.pres_time = (fsi.pres_time << 8) + tag.Timestamp[2];

	if (key) fsi.size |= FLV_KEYFRAME;

	dv->Add(&fsi);
	ret = true;

Abort:
	return ret;
}


#ifdef FLV_VFR

void VideoTrackFLV::DeriveFrameRate()
{
	double Duration;
	double msperframe;
	double min;

	unsigned long min_pts;
	unsigned long max_pts;
	unsigned long min_diff;
	unsigned long max_diff;
	unsigned long i, t;

	unsigned long histo[500] = { 0 };

	// Start analysis...
	// The most difficult thing here is choosing a good output
	// frame rate; after we have decided on the output frame
	// rate, the rest is fairly straightforward.

	// Compute histogram of differences in presentation time.
	// While we're at it, get the min and max presentation times
	// and the min and max presentation time differences.

	min_pts  = sample_list[0].pres_time;
	min_diff = sample_list[1].pres_time - min_pts;
	max_pts  = min_pts;
	max_diff = min_diff;

	for (i = 1; i < dwActualLength; ++i)
	{
		unsigned long diff = sample_list[i].pres_time;

		if (diff < min_pts) min_pts = diff;
		if (diff > max_pts) max_pts = diff;

		diff -= sample_list[i - 1].pres_time;
		if (diff != 0)
		{
			if (diff < min_diff) min_diff = diff;
			if (diff > max_diff) max_diff = diff;

			if (diff <= 500)
			{
				++histo[diff - 1];
			}
		}
	}

	// Calculate the true duration in seconds as follows:
	// msperframe = (max_pts - min_pts) / (dwActualLength - 1)
	// Duration   = (msperframe / 1000) * dwActualLength
	
	Duration  = max_pts - min_pts;
	Duration *= dwActualLength;
	Duration /= 1000.0 * (dwActualLength - 1);

	// Pick out the first frame difference that occurs for at least
	// 10% of the video.  Ignore any timings below 10ms (100fps).

	t = (dwActualLength + 5) / 10;		// threshold

	// Note that "threshold" will be zero if length is less
	// than 5.  We do not want histo[i] to be zero!

	for (i = 10; i < 500; ++i)
	{
		if (histo[i] > t) break;
	}

	// Remember that diff was stored in slot[diff - 1]!

	// Now we have histo[i] frames with diff (i + i) as a base.
	// Add any other rates which are +/-1 ms from (i + 1).

	if (i < 500)
	{
		double diff_cnt = histo[i];
		msperframe = (double)histo[i] * (i + 1);
		
		if (i > 10)
		{
			diff_cnt   += histo[i - 1];
			msperframe += (double)histo[i - 1] * (i + 0);
		}
		if (i < 499)
		{
			diff_cnt   += histo[i + 1];
			msperframe += (double)histo[i + 1] * (i + 2);
		}

		msperframe /= diff_cnt;
	}
	else
	{
		// No rate was used for at least 10% of the movie?
		// In that case, just settle for a wild guess
		msperframe  = max_pts - min_pts;
		msperframe /= (dwActualLength - 1);
	}

	// Let's try to lock the rate to something exact:

	static const struct {
		DWORD mRate;
		DWORD mScale;
	} rates[12] = {
		{60000, 1000},	// 60		60p
		{60000, 1001},	// 59.94	NTSC double rate
		{30000, 1000},	// 30		30p
		{30000, 1001},	// 29.97	NTSC
		{25000, 1000},	// 25		PAL/SECAM
		{24000, 1000},	// 24		Film
		{24000, 1001},	// 23.976	NTSC film
		{18000, 1000},	// 18		Silent film
		{18000, 1001},	// 17.982	NTSC silent film
		{16000, 1000},	// 16		Silent film
		{16000, 1001},	// 15.984	NTSC silent film
		{15000, 1000}	// 15		Half 30p
	};

	// Find the rate with the least amount
	// of deviation (in ms per frame):

	t = 0;
	min = 1.0;

	for (i = 0; i < 12; ++i)
	{
		double diff = msperframe - (
			(1000.0 * rates[i].mScale) / rates[i].mRate
		);

		if (diff < 0) diff = -diff;

		if (diff < min)
		{
			min = diff;
			t = i + 1;		// closest so far
		}
	}

	// Finally, let's accumulate the deviation for the duration
	// of the movie, and arbitrarily say that we'll allow
	// no more than 10ms of deviation for that whole period.

	// The amount of ms we might lose by locking is:
	// deviation_in_ms_per_frame * fps * duration_in_seconds

	min *= 1000.0 * Duration;
	min /= msperframe;

	if (t > 0 && min < 10.0)
	{
		// Awesome; lock to this
		dwRate  = rates[t - 1].mRate;
		dwScale = rates[t - 1].mScale;
	}
	else
	{
		// Measured rate is too far off; don't lock.
		// Instead write it as microseconds per frame
		dwRate  = 1000000000;
		dwScale = (DWORD)(msperframe * 1000000.0 + 0.5);
	}

	// No mapping is needed if (max_diff - min_diff) is one.
	// In this case just play what we have at the chosen rate.

	if ((max_diff - min_diff) == 1)
	{
		// Derive constant frame rate map
		dwPerceivedLength = dwActualLength;
		sample_map = new unsigned long[dwActualLength];
		if (sample_map == NULL) goto Abort;
		for (i = 0; i < dwActualLength; ++i) sample_map[i] = i;
	}
	else
	{
		// Derive variable frame rate map.
		// Start by determining how many frames we must present.

		// This will drift a bit where pts differences are less than
		// the chosen rate, but we already know that these differences
		// account for less than 10% of the video, so hopefully the
		// map will catch up before too much sync is lost.

		DWORD dwSampleIn = 0;
/*
		for (dwPerceivedLength = 0; (dwSampleIn + 1) < dwActualLength; ++dwPerceivedLength)
		{
			// Calculate the ideal presentation time of frame i + one half
			t = (DWORD)(((0.5 + dwPerceivedLength) * 1000.0 * dwScale) / dwRate + 0.5);

			if (t >= sample_list[dwSampleIn + 1].pres_time)
			{
				++dwSampleIn;
			}
		}
*/
		for (dwPerceivedLength = 0; (dwSampleIn + 1) < dwActualLength; )
		{
			// Calculate the ideal presentation time of frame i + one half
			t = (DWORD)(((0.5 + dwPerceivedLength) * 1000.0 * dwScale) / dwRate + 0.5);

			if (t < sample_list[dwSampleIn + 1].pres_time)
			{
				++dwPerceivedLength;
				continue;
			}
			++dwSampleIn;
		}

		// Now generate the mapping from perceived frame to media sample.

		sample_map = new unsigned long[dwPerceivedLength];
		if (sample_map == NULL) goto Abort;

		dwSampleIn = 0;
/*
		for (i = 0; i < dwPerceivedLength; ++i)
		{
			// Calculate the ideal presentation time of frame i + one half
			t = (DWORD)(((0.5 + i) * 1000.0 * dwScale) / dwRate + 0.5);

			if (dwSampleIn < (dwActualLength - 1))
			{
				if (t >= sample_list[dwSampleIn + 1].pres_time)
				{
					++dwSampleIn;
				}
			}
			sample_map[i] = dwSampleIn;
		}
*/
		for (i = 0; i < dwPerceivedLength; )
		{
			// Calculate the ideal presentation time of frame i + one half
			t = (DWORD)(((0.5 + i) * 1000.0 * dwScale) / dwRate + 0.5);

			if (t < sample_list[dwSampleIn + 1].pres_time)
			{
				sample_map[i++] = dwSampleIn;
				continue;
			}
			++dwSampleIn;
		}
	}

Abort:

#if 0
	{
		FILE *file = fopen("E:\\stamps.txt", "w");
		if (file != NULL)
		{
			for (i = 0; i < dwActualLength; ++i)
			{
				if (i > 0)
				{
					fprintf(file, "%lu (%lu)\n", sample_list[i].pres_time,
						sample_list[i].pres_time - sample_list[i-1].pres_time
					);
				}
				else
				{
					fprintf(file, "%lu\n", sample_list[i].pres_time);
				}
			}
			fclose(file);
		}
	}
#endif

	return;
}

#endif	// FLV_VFR


bool VideoTrackFLV::IsDroppedFrame(unsigned long frame) const
{
	if (frame != 0)
	{
		if (frame < dwPerceivedLength)
		{
			return (sample_map[frame] == sample_map[frame - 1]);
		}
		return true;
	}
	return false;
}


bool VideoTrackFLV::IsKey(unsigned long frame) const
{
	if (frame < dwPerceivedLength && !IsDroppedFrame(frame))
	{
		return ((sample_list[sample_map[frame]].size & FLV_KEYFRAME) != 0);
	}
	return false;
}


bool VideoTrackFLV::Finish()
{
	unsigned long i;

	if (dv == NULL) goto Abort;

	sample_list = (FLVSampleInfo *)dv->MakeArray();
	dwActualLength = dv->Length();

	delete dv;
	dv = NULL;

	if (dwActualLength == 0) goto Abort;

	// Beware of VP6 file without any keyframes
	if (bmih.biWidth == 0 && bmih.biHeight == 0) goto Abort;

#ifdef FLV_VFR
	if (dwActualLength > 2)
	{
		DeriveFrameRate();
	}
	else if (dwActualLength == 2)
	{
		dwRate  = 1000;
		dwScale = sample_list[1].pres_time - sample_list[0].pres_time;
	}
	else
	{
		// Only one frame!
		dwRate  = 1000;
		dwScale = 40;
	}
/*
	i = dwActualLength;
	if (i >= 2) {
		long lFirst = sample_list[0].pres_time;
		long lLast = sample_list[i - 1].pres_time;

		double fr =
			((double)dwRate * (double)(lLast - lFirst))
			/ ((double)dwScale * 1000.0);

		dwPerceivedLength = (uint32)(fr + 0.5) + 1;
	}
*/
#else
	i = dwActualLength;
	if (i > 1)
	{
		DWORD min_pts = sample_list[0].pres_time;
		DWORD max_pts = min_pts;
		double msperframe;

		for (i = 1; i < dwActualLength; ++i)
		{
			DWORD pts = sample_list[i].pres_time;
			if (pts < min_pts) min_pts = pts;
			if (pts > max_pts) max_pts = pts;
		}

		msperframe  = max_pts - min_pts;
		msperframe /= dwActualLength - 1;

		dwRate = 1000000000;
		dwScale = (DWORD)(msperframe * 1000000.0 + 0.5);
	}
	else
	{
		// Only one frame!
		dwRate  = 1000;
		dwScale = 40;
	}

	// Create sample map
	dwPerceivedLength = dwActualLength;
	sample_map = new unsigned long[dwActualLength];
	if (sample_map == NULL) goto Abort;
	for (i = 0; i < dwActualLength; ++i) sample_map[i] = i;
#endif

	return true;

Abort:
	return false;
}


///////////////////////////////////////////////////////////////////////
//
//                      VideoSourceFLV
//
///////////////////////////////////////////////////////////////////////


VideoSourceFLV::VideoSourceFLV(InputFileFLV *const pp, VideoTrackFLV *const vt)
	: parentPtr(pp)
	, video(vt)
{
	unsigned long max;
	unsigned long i;

	parentPtr->AddRef();

	errmode = kErrorModeDecodeAnyway;

	dwPrevSample = -1;
	dwGuess = 0;

	max = 0;
	for (i = 0; i < video->dwActualLength; ++i) {
		unsigned long sz = video->sample_list[i].size & (~FLV_KEYFRAME);
		if (sz > max) max = sz;
	}
	if (max < 4) max = 4;
	frame_buffer = new unsigned char[max];

_RPT0(_CRT_WARN, "VideoSourceFLV constructed\n");
}


VideoSourceFLV::~VideoSourceFLV()
{
	delete[] frame_buffer;

	parentPtr->Release();

_RPT0(_CRT_WARN, "VideoSourceFLV destructed\n");
}


int VideoSourceFLV::AddRef()
{
_RPT0(_CRT_WARN, "VideoSourceFLV::AddRef\n");
	return vdxunknown<IVDXStreamSource>::AddRef();
}


int VideoSourceFLV::Release()
{
_RPT0(_CRT_WARN, "VideoSourceFLV::Release\n");
	return vdxunknown<IVDXStreamSource>::Release();
}


void *VDXAPIENTRY VideoSourceFLV::AsInterface(uint32 iid)
{
	if (iid == IVDXVideoSource::kIID)
		return static_cast<IVDXVideoSource *>(this);

	return vdxunknown<IVDXStreamSource>::AsInterface(iid);
}


void VDXAPIENTRY VideoSourceFLV::GetStreamSourceInfo(VDXStreamSourceInfo& srcInfo)
{
	srcInfo.mSampleRate.mNumerator = video->dwRate;
	srcInfo.mSampleRate.mDenominator = video->dwScale;
	srcInfo.mSampleCount = video->dwPerceivedLength;

	// TODO:
	srcInfo.mPixelAspectRatio.mNumerator = 1;
	srcInfo.mPixelAspectRatio.mDenominator = 1;
}


int VideoSourceFLV::UnwrapAVCFrame(unsigned char *dst, const unsigned char *src, int len)
{
	// Unwrap NALUs
	int avc_len = video->avc_len;
	int ret = 0;

	while (len > avc_len)
	{
		int size = *src++;
		--len;
		if (avc_len > 1) {
			size = (size << 8) + *src++;
			--len;
			if (avc_len > 2) {
				size = (size << 8) + *src++;
				--len;
				if (avc_len > 3) {
					size = (size << 8) + *src++;
					--len;
				}
			}
		}
		if (size > len) size = len;

		if (dst != NULL) {
			*dst++ = 0;
			*dst++ = 0;
			*dst++ = 0;
			*dst++ = 1;
			memcpy(dst, src, size);
			dst += size;
		}

		ret += 4 + size;
		src += size;
		len -= size;
	}

	return ret;
}


bool VideoSourceFLV::Read(sint64 lStart64, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead)
{
	unsigned char *dst = (unsigned char *)lpBuffer;
	int framesize;
	int actual;

	DWORD dwStart = (DWORD)lStart64;

	if (dwStart >= video->dwPerceivedLength)
	{
		if (lBytesRead) *lBytesRead = 0;
		if (lSamplesRead) *lSamplesRead = 0;
		return true;
	}

	// Dropped frame?
	if (video->IsDroppedFrame(dwStart))
	{
		if (lBytesRead) *lBytesRead = 0;
		if (lSamplesRead) *lSamplesRead = 1;
		return true;
	}
	dwStart = video->sample_map[dwStart];

	framesize = video->sample_list[dwStart].size & (~FLV_KEYFRAME);

//_RPT2(_CRT_WARN, "VideoSourceFLV::Read %i, %i\n", lStart, dwPrevSample);

	if (video->bFormat == FLV_VIDEO_AVC) {
		// read data, but only if it's new data
		if (dwStart != dwPrevSample) {
			dwPrevSample = dwStart;
			parentPtr->ReadData(frame_buffer, video->sample_list[dwStart].pos, framesize);
		}
		actual = UnwrapAVCFrame(NULL, frame_buffer, framesize);
		if (video->sample_list[dwStart].size & FLV_KEYFRAME) actual += video->dcr_size;
	} else {
		actual = framesize;
	}

	if (dst == NULL || cbBuffer < (uint32)actual)
	{
		if (lBytesRead) *lBytesRead = actual;
		if (lSamplesRead) *lSamplesRead = 1;
		return (dst == NULL);
	}

	if (video->bFormat == FLV_VIDEO_AVC)
	{
		if (video->sample_list[dwStart].size & FLV_KEYFRAME) {
			memcpy(dst, video->avc_dcr, video->dcr_size);
			dst += video->dcr_size;
		}
		dst += UnwrapAVCFrame(dst, frame_buffer, framesize);
	}
	else
	{ 
		// read data, but only if it's new data
		if (dwStart != dwPrevSample) {
			dwPrevSample = dwStart;
			parentPtr->ReadData(frame_buffer, video->sample_list[dwStart].pos, framesize);
		}
		memcpy(dst, frame_buffer, framesize);
		dst += framesize;
	}

	if (lBytesRead) *lBytesRead = (uint32)(dst - (unsigned char *)lpBuffer);
	if (lSamplesRead) *lSamplesRead = 1;
	return true;
}


const void *VideoSourceFLV::GetDirectFormat()
{
	return &video->bmih;
}


int VideoSourceFLV::GetDirectFormatLen()
{
	return sizeof(video->bmih);
}


IVDXStreamSource::ErrorMode VideoSourceFLV::GetDecodeErrorMode()
{
	return errmode;
}


void VideoSourceFLV::SetDecodeErrorMode(IVDXStreamSource::ErrorMode mode)
{
	errmode = mode;
}


bool VideoSourceFLV::IsDecodeErrorModeSupported(IVDXStreamSource::ErrorMode mode)
{
	return true;
}


bool VideoSourceFLV::IsVBR()
{
	return false;
}


sint64 VideoSourceFLV::TimeToPositionVBR(sint64 us)
{
	return 0;
}


sint64 VideoSourceFLV::PositionToTimeVBR(sint64 samples)
{
	return 0;
}


void VideoSourceFLV::GetVideoSourceInfo(VDXVideoSourceInfo& info)
{
	info.mWidth  = video->bmih.biWidth;
	info.mHeight = video->bmih.biHeight;
	info.mFlags  = 0;

#ifdef USE_DEFAULT_IP_MODEL
	info.mDecoderModel = VDXVideoSourceInfo::kDecoderModelDefaultIP;
#else
	info.mDecoderModel = VDXVideoSourceInfo::kDecoderModelCustom;
#endif
}


bool VideoSourceFLV::CreateVideoDecoderModel(IVDXVideoDecoderModel **ppModel)
{
#ifdef USE_DEFAULT_IP_MODEL
	*ppModel = NULL;
	return false;
#else
	*ppModel = new VideoDecoderModelFLV(this, video);
	if (*ppModel == NULL) return false;

	(*ppModel)->AddRef();
	return true;
#endif
}


bool VideoSourceFLV::CreateVideoDecoder(IVDXVideoDecoder **ppDecoder)
{
#ifdef DEFER_DECODING
	*ppDecoder = NULL;
	return false;
#else
	*ppDecoder = new VideoDecoderFLV(this, video);
	if (*ppDecoder == NULL) return false;

	(*ppDecoder)->AddRef();
	return true;
#endif
}


void VideoSourceFLV::GetSampleInfo(sint64 sample_num, VDXVideoFrameInfo& frameInfo)
{
	DWORD dwSample = (DWORD)sample_num;

	if (dwSample < video->dwPerceivedLength)
	{
		if (video->IsDroppedFrame(dwSample))
		{
			frameInfo.mFrameType    = kVDXVFT_Null;
			frameInfo.mTypeChar     = 'D';
			frameInfo.mBytePosition = -1;
		}
		else
		{
			const FLVSampleInfo& fsi =
				video->sample_list[video->sample_map[dwSample]];

			if (fsi.size & FLV_KEYFRAME)
			{
				frameInfo.mFrameType = kVDXVFT_Independent;
				frameInfo.mTypeChar  = 'K';
			}
			else
			{
				frameInfo.mFrameType = kVDXVFT_Predicted;
				frameInfo.mTypeChar  = ' ';
			}
			frameInfo.mBytePosition  = fsi.pos;
		}
	}
	else
	{
		// sample_num out of bounds
		frameInfo.mFrameType    = kVDXVFT_Null;
		frameInfo.mTypeChar     = 'M';
		frameInfo.mBytePosition = -1;
	}
}


bool VideoSourceFLV::IsKey(sint64 sample)
{
	return video->IsKey((unsigned long)sample);
}


sint64 VideoSourceFLV::GetFrameNumberForSample(sint64 sample_num)
{
	return sample_num;
}


sint64 VideoSourceFLV::GetSampleNumberForFrame(sint64 display_num)
{
	return display_num;
}


sint64 VideoSourceFLV::GetRealFrame(sint64 display_num)
{
	return display_num;
}


sint64 VideoSourceFLV::GetSampleBytePosition(sint64 sample_num)
{
	DWORD dwSample = (DWORD)sample_num;

	if (dwSample < video->dwPerceivedLength)
	{
		if (!video->IsDroppedFrame(dwSample))
		{
			return video->sample_list[video->sample_map[dwSample]].pos;
		}
	}
	return -1;
}


