///////////////////////////////////////////////////////////////////////
// FLV plugin for VirtualDub 1.8.8+
// Copyright (C) 2010-2012 fccHandler
// Copyright (C) 1998-2012 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

// Windows file I/O functions, wrapped in a class
// Files larger than 4GB are supported on NTFS.

#include <windows.h>
#include <memory.h>

#include "WinFile.h"


WinFile::WinFile()
{
	m_hFile         = INVALID_HANDLE_VALUE;
	m_Pos.QuadPart  = 0i64;
	m_Size.QuadPart = 0i64;
	m_CurByte       = 0;
	m_CurBit        = 0;

	memset(m_Buffer, 0, WF_BUFFERSIZE);
}


WinFile::~WinFile()
{
	if (m_hFile != INVALID_HANDLE_VALUE) CloseHandle(m_hFile);
}


/*
	How Windows handles some aberrant situations:

	Setting the file pointer to absolute -10 returns 0xFFFFFFFF
	and error code 87 (ERROR_INVALID_PARAMETER), and does
	not move the file pointer (apparently)

	Setting the file pointer to relative -10 returns 0xFFFFFFFF
	and error code 131 (ERROR_NEGATIVE_SEEK), and does
	not move the file pointer

	Setting the file pointer AT or BEYOND the end of file succeeds.

	Attempting to read past the end of file does not cause
	an error, but reads zero bytes.
*/

void WinFile::Refresh()
{
	DWORD dwHi = m_Pos.HighPart;
	DWORD dwLo = m_Pos.LowPart & -WF_BUFFERSIZE;

	SetFilePointer(m_hFile, dwLo, (PLONG)&dwHi, FILE_BEGIN);

	if (ReadFile(m_hFile, m_Buffer, WF_BUFFERSIZE, &dwHi, NULL) == 0)
	{
		dwHi = 0;
	}

	if (dwHi < WF_BUFFERSIZE)
	{
		memset(m_Buffer + dwHi, 0, WF_BUFFERSIZE - dwHi);
	}
}


bool WinFile::Open(HANDLE h)
{
	if (h != INVALID_HANDLE_VALUE)
	{
		DWORD dwHi = 0;
		DWORD dwLo = GetFileSize(h, &dwHi);

		if (dwLo != INVALID_FILE_SIZE || GetLastError() == NO_ERROR)
		{
			if (dwHi < 0x80000000U)
			{
				if (m_hFile != INVALID_HANDLE_VALUE)
				{
					CloseHandle(m_hFile);
				}

				m_hFile         = h;
				m_Pos.QuadPart  = 0;
				m_Size.LowPart  = dwLo;
				m_Size.HighPart = dwHi;
				m_CurByte       = 0;
				m_CurBit        = 0;

				Refresh();
				return true;
			}
		}

		CloseHandle(h);
	}

	return false;
}


bool WinFile::OpenA(const char *pFileName)
{
	if (pFileName != NULL)
	{
		HANDLE h = CreateFileA(pFileName, GENERIC_READ, FILE_SHARE_READ,
			NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);

		return Open(h);
	}

	return false;
}


bool WinFile::OpenW(const wchar_t *pFileName)
{
	HANDLE h = INVALID_HANDLE_VALUE;

	if (pFileName != NULL)
	{
		if (GetVersion() & 0x80000000U)
		{
			// Windows 95/98
			int len = WideCharToMultiByte(CP_ACP, 0, pFileName, -1, NULL, 0, NULL, NULL);

			if (len > 1)
			{
				char *pFileNameA = new char[len + 2];

				if (pFileNameA != NULL)
				{
					if (WideCharToMultiByte(CP_ACP, 0, pFileName, -1, pFileNameA, len + 1, NULL, NULL) == len)
					{
						h = CreateFileA(pFileNameA, GENERIC_READ, FILE_SHARE_READ,
							NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
					}
					
					delete[] pFileNameA;
				}
			}
		}
		else
		{
			// Windows NT/XP
			h = CreateFileW(pFileName, GENERIC_READ, FILE_SHARE_READ,
				NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
		}
	}

	return Open(h);
}


bool WinFile::Seek(__int64 iAbsPos)
{
	LARGE_INTEGER OldPos;
	bool ret = false;

	if (m_hFile == INVALID_HANDLE_VALUE) goto Abort;
	if (iAbsPos < 0) goto Abort;

	ret = true;
	if (iAbsPos > m_Size.QuadPart)
	{
		iAbsPos = m_Size.QuadPart;
		ret = false;
	}

	OldPos.QuadPart = m_Pos.QuadPart & -WF_BUFFERSIZE;
	m_Pos.QuadPart  = iAbsPos;

	if ((m_Pos.QuadPart & -WF_BUFFERSIZE) != OldPos.QuadPart)
	{
		Refresh();
	}

	m_CurByte = 0;
	m_CurBit  = 0;

Abort:
	return ret;
}


long WinFile::Read(void *pBuffer, int iBytes)
{
	int remains, offs;
	unsigned char *dst;

	long ret = 0;

	if (m_hFile == INVALID_HANDLE_VALUE) goto Abort;
	if (pBuffer == NULL) goto Abort;
	if (iBytes <= 0) goto Abort;

	m_CurByte = 0;
	m_CurBit = 0;

	// Don't read past EOF
	if ((m_Pos.QuadPart + (__int64)iBytes) > m_Size.QuadPart)
	{
		if (m_Pos.QuadPart >= m_Size.QuadPart) goto Abort;
		iBytes = (int)(m_Size.QuadPart - m_Pos.QuadPart);
	}

	dst = (unsigned char *)pBuffer;

	offs = m_Pos.LowPart & (WF_BUFFERSIZE - 1);
	remains = WF_BUFFERSIZE - offs;

	for (;;)
	{
		if (remains > iBytes) remains = iBytes;
		memcpy(dst, m_Buffer + offs, remains);
		ret += remains;
		m_Pos.QuadPart += remains;
		iBytes -= remains;
		if (iBytes <= 0) break;
		dst += remains;
		Refresh();
		offs = 0;
		remains = WF_BUFFERSIZE;
	}

	if ((m_Pos.LowPart & (WF_BUFFERSIZE - 1)) == 0) Refresh();

Abort:
	return ret;
}


int WinFile::NextByte()
{
	int ret = 0;

	if (m_hFile != INVALID_HANDLE_VALUE)
	{
		if (m_Pos.QuadPart < m_Size.QuadPart)
		{
			int offs = m_Pos.LowPart & (WF_BUFFERSIZE - 1);
			ret = m_Buffer[offs];
			++m_Pos.QuadPart;
			if (offs == (WF_BUFFERSIZE - 1)) Refresh();
		}
	}

	return ret;
}


void WinFile::AlignToByte()
{
	m_CurByte = 0;
	m_CurBit  = 0;
}


int WinFile::NextBit()
{
	int ret;

	if (m_CurBit == 0)
	{
		m_CurByte = NextByte();
		m_CurBit = 128;
	}

	ret = (m_CurByte & m_CurBit)? 1: 0;
	m_CurBit >>= 1;

	return ret;
}


int WinFile::GetBits(int nBits)
{
	int ret = 0;

	while (nBits > 0)
	{
		ret = ret + ret + NextBit();
		--nBits;
	}

	return ret;
}

